#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define SECURE_ROUNDS 16 // Wild guess
#define WIDTH 32
#define STATESPERBLOCK 32
#define BLOCKLEN (STATESPERBLOCK*WIDTH) // 4KiB

static inline uint32_t rotate(uint32_t value, uint32_t dist) {
    return (value << dist) | (value >> (32 - dist));
}

static void printState(uint32_t *state) {
/*
    for(uint32_t i = 0; i < WIDTH; i++) {
        for(uint32_t j = 0; j < WIDTH; j++) {
            if(state[i] & ((uint32_t)1 << j)) {
                printf("X");
            } else {
                printf(" ");
            }
        }
        printf("\n");
    }
    printf("\n");
*/
}

// Use Wolfram's Rule 30 to hash all 32 rows.  Also do a diagonalization.
static inline void hashRows(uint32_t *state, uint32_t numRounds) {
    for(uint32_t i = 0; i < numRounds; i++) {
        for(uint32_t j = 0; j < WIDTH; j++) {
            state[j] ^= rotate(state[j], 1) | rotate(state[j], 2);
        }
    }
}

// Use Wolfram's Rule 30 to hash all 32 columns.
static inline void hashColumns(uint32_t *state, uint32_t numRounds) {
    for(uint32_t i = 0; i < numRounds; i++) {
        uint32_t state0 = state[0];
        uint32_t state1 = state[1];
        for(uint32_t j = 0; j < WIDTH-2; j++) {
            state[j] ^= state[j+1] | state[j+2];
        }
        state[WIDTH-2] ^= state[WIDTH-1] | state0;
        state[WIDTH-1] ^= state0 | state1;
    }
}

static inline void diagonalize(uint32_t *state) {
    for(uint32_t j = 0; j < WIDTH; j++) {
        state[j] = rotate(state[j], j);
    }
}

// Apply hash rounds to mix the state.
static void hashRounds(uint32_t *state, uint32_t numRounds) {
    printState(state);
    for(uint32_t i = 0; i < numRounds; i++) {
        hashColumns(state, SECURE_ROUNDS);
        printState(state);
        diagonalize(state);
        printState(state);
        hashRows(state, SECURE_ROUNDS);
        printState(state);
    }
}

// Suck a uint32 into the state, and apply rounds.
static inline void absorb(uint32_t *state, uint32_t value) {
    state[0] += value;
    hashRounds(state, SECURE_ROUNDS);
}

// Suck in a uint8 array.
static void absorbUint8Array(uint32_t *state, const uint8_t *in, uint32_t inlen) {
    while(inlen >= 4) {
        uint32_t value = in[0] | (((uint32_t)in[1]) << 8) | (((uint32_t)in[2]) << 16) | (((uint32_t)in[3]) << 16);
        absorb(state, value);
        inlen -= 4;
        in += 4;
    }
    if(inlen > 0) {
        uint8_t buf[4];
        memcpy(buf, in, inlen);
        absorbUint8Array(state, buf, inlen);
    }
}

// Write an array of uint32s, and apply rounds every time we've written half the state.
static void squeeze(uint32_t *state, uint32_t *dest, uint32_t length, uint32_t numRounds) {
    while(length >= WIDTH/2) {
        hashRounds(state, numRounds);
        memcpy(dest, state, WIDTH/2*sizeof(uint32_t));
        dest += WIDTH/2;
        length -= WIDTH/2;
    }
    if(length > 0) {
        hashRounds(state, numRounds);
        memcpy(dest, state, length*sizeof(uint32_t));
    }
}

// Write an array of uint8s, and apply rounds every time we've written half the state.
static void squeezeUint8(uint32_t *state, uint8_t *dest, uint32_t length) {
    while(length > 0) {
        for(uint32_t i = 0; i < WIDTH/2 && length > 0; i++) {
            uint32_t a = state[i];
            for(uint32_t j = 0; j < 4 && length > 0; j++) {
                *dest++ = (uint8_t)a;
                a >>= 8;
                length--;
            }
        }
        hashRounds(state, SECURE_ROUNDS);
    }
}

// This is only used in mixBlocks.  It does not apply any hashing.
static inline void absorb2(uint32_t *state, uint32_t *in1, uint32_t *in2) {
    for(uint32_t i = 0; i < WIDTH; i++) {
        state[i] = *in1++ + *in2++;
    }
}

// Hash two blocks together, and write the result to the next block in memory.
static void mixBlocks(uint32_t *mem, uint32_t *state, uint64_t sourceAddr, uint64_t destAddr) {
    uint32_t *in = mem + sourceAddr;
    uint32_t *out = mem + destAddr;
    uint32_t *carry = out - BLOCKLEN;
    for(uint32_t i = 0; i < STATESPERBLOCK; i += 2) {
        absorb2(state, in, carry);
        hashColumns(state, 1);
        memcpy(out, state, WIDTH*sizeof(uint32_t));
        in += WIDTH;
        out += WIDTH;
        carry += WIDTH;
        absorb2(state, in, carry);
        hashRows(state, 1);
        memcpy(out, state, WIDTH*sizeof(uint32_t));
        in += WIDTH;
        out += WIDTH;
        carry += WIDTH;
    }
    diagonalize(state);
}

// The main hashing function.
int gpuHash(void *out, size_t outlen, const void *in, size_t inlen, const void *salt,
        size_t saltlen, unsigned int m_cost) {
    uint32_t state[WIDTH] = {0,};
    uint32_t numBlocks = 1 << m_cost; // Each block is 4KiB

    absorb(state, saltlen);
    absorbUint8Array(state, salt, saltlen);
    absorb(state, inlen);
    absorbUint8Array(state, in, inlen);
    absorb(state, outlen);
    absorb(state, m_cost);

    //printf("Allocating %lu bytes\n", ((uint64_t)BLOCKLEN << m_cost)*sizeof(uint32_t));
    uint32_t *mem = calloc((uint64_t)BLOCKLEN << m_cost, sizeof(uint32_t));
    squeeze(state, mem, 2*BLOCKLEN, 1);
    for(uint32_t i = 2; i < numBlocks; i++) {
        uint32_t randBlock = (state[0] % (i-1))*BLOCKLEN;
        mixBlocks(mem, state, randBlock, i*(uint64_t)BLOCKLEN);
    }
    //printState(state);
    squeezeUint8(state, out, outlen);
    return 0;
}

int PHS(void *out, size_t outlen, const void *in, size_t inlen, const void *salt,
        size_t saltlen, unsigned int t_cost, unsigned int m_cost) {
    return gpuHash(out, outlen, in, inlen, salt, saltlen, m_cost);
}
